// All rights reserved by Daisuke Fukuoka and Takeshi Hara @ Gifu University. 
// 2002, 2003, 2004, 2005

#include <stdlib.h>
#include <stdio.h>
#include <sys/stat.h>
#include <GL/glut.h>


#define USAGE "Usages :\n\
xctview [filename][width][height][wl][ww][HU offset][header(byte)][(no)swap]\n\
\n\
[comments]\n\
filename\t: Image data you desire to open.\n\
width\t\t: Image width [# of pixel]\n\
height\t\t: Image height [# of pixel]\n\
wl\t\t: Window level to display [-3000, 3000]\n\
ww\t\t: Window width to display [0, 6000]\n\
HU offset\t: If HU values were biased, indicate the value.\n\
header\t\t: # of bytes in header area of the file\n\
(no)swap\t: swap the file for little/big endian. Indicate noswap or swap.\n\
slab depth\t\t: # of slices to MIP display\n\n\
[KEY ASSIGN]\n\
a/d\t: -/+ WL\n\
x/s\t: -/+ WW\n\
n/b\t: slice # +/-\n\
t/w\t: jump to start/end\n\
f/v\t: -/+ step [default: 20 HU]\n\
</>\t: -/+ MIP depth [default: 10 images]\n\
r\t: back to start slice and reset WL/WW\n\
e\t: jump to end slice and reset WL/WW\n\
l\t: CT pulmonary [WL WW]=[-595, 842]\n\
k\t: CT abdomen   [WL WW]=[  40, 350]\n\
j\t: CT bone      [WL WW]=[ 300,1500]\n\
h\t: CT brain     [WL WW]=[  50, 100]\n\
q\t: quit\n\
\n"

#define HUmax	3000
#define HUmin	-3000

#define CTpulWL	-595
#define CTpulWW	 842
#define CTabdWL	  40
#define CTabdWW	 350
#define CTbonWL  300
#define CTbonWW	1500
#define CTbraWL   50
#define CTbraWW	 100


/*
#define Scale 1.00
*/

short *ctdata,*maximg, *minimg;
unsigned char *image;
char *fname;

void _init(void);
void _display(void);
void _reshape(int,int);
void _read_image(char *);
int  _file_size(char *);
void _get_args(int argc, char **argv);
void _create_display_image(int wl, int ww);
void _create_display_image_of_n(int wl, int ww, int num);
void byte_swap_short(short *in, int width,int height);
unsigned char win_proc(int value, int level, int width);
void vol2mip(int num, int depth, int wl, int ww);

void function(short *ctimg, int width, int height, int slicenum, int threshold);

int	win_height, Height;
int win_width, Width;
int Slice;
int n_slice=0;//n_sliceɽƤ륹饤ֹ
int offset,huoffset;
int wlevel, wwidth,start_slice, num_of_slice,orgwl,orgww;
int hustep=20;
int swap=1;
int slabdepth=10; //MIPɽѥָ
int threshold; //齬ӤΤ

double Scale;

GLenum doubleBuffer;

struct stat stat_buf;

void _read_image(char *filename){
  int i,j;

  FILE *fp;
	
	doubleBuffer = GL_TRUE;

  if((fp=fopen(filename,"rb"))==NULL){ 
    printf("Error:Can't Read File\n");
    exit(-1);
  }            
	fseek(fp,(long)offset,0);	
	fread(ctdata,sizeof(short),Width*Height*Slice,fp);
    fclose(fp);
	
	printf("Filename: %s\n",filename);
}

void _GL_init(void){
   glClearColor (0.0, 0.0, 0.0, 0.0);
   glShadeModel(GL_FLAT);
   glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
}

void _GL_display(void){
   glClear(GL_COLOR_BUFFER_BIT);
   glRasterPos2i(0,win_height);

   glPixelZoom((double)win_width/(double)Width,
	       -(double)win_height/(double)Height);

   glDrawPixels(Width,Height,GL_LUMINANCE,GL_UNSIGNED_BYTE,image);

   glutSwapBuffers();
   /*  
   glFlush();
   */
}

void _GL_reshape(int w, int h){

  win_height=h;  win_width=w;

  glViewport(0, 0, (GLsizei) w, (GLsizei) h);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluOrtho2D(0.0, (GLdouble) w, 0.0, (GLdouble) h);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
}

void _get_args(int argc, char **argv)
{

  if(9 != argc){
    fprintf(stderr,USAGE);
    exit(1);
  }

  Scale =1.0;
	offset = 0;
	
	stat(argv[1],&stat_buf);
	fname = argv[1];
	Width = atoi(argv[2]); printf("\nW = %d,", Width);
	Height = atoi(argv[3]); printf("H = %d,", Height);
	wlevel = atoi(argv[4]); printf("WL = %d,", wlevel);orgwl=wlevel;
	wwidth = atoi(argv[5]); printf("WW = %d,", wwidth);orgww=wwidth;
	huoffset = atoi(argv[6]); printf("\nHU offset = %d,", huoffset);
	offset = atoi(argv[7]); printf("header size[byte] = %d,", offset);
	swap = strcmp("swap", argv[8]); //printf("swap = %d,", swap);
	
	//饤μưꡥstatȤ
	Slice = stat_buf.st_size/Width/Height/sizeof(short); printf("\n# of slice = %d,", Slice);
}

static void Key(unsigned char key, int x, int y)
{
	int i;
    (void) x; 
    (void) y;
    switch (key) {
      case 27:
		exit(1);
		
	  case'g':
		function(ctdata, Width, Height, Slice, threshold);
		vol2mip(n_slice, slabdepth, wlevel, wwidth);
		glutPostRedisplay();
		return;
		
	  case 'd':
		if(wlevel<=HUmax+hustep+1){
			wlevel+=hustep;
		}
		printf("(Window Level, Window Width, # of slice)=(%d, %d, %d)\n",wlevel, wwidth, n_slice+1);
	_create_display_image_of_n(wlevel,wwidth,n_slice);
		glutPostRedisplay();
		return;
		
      case 'a':
		if(wlevel>=HUmin-hustep-1){
			wlevel-=hustep;
		}
		printf("(Window Level, Window Width, # of slice)=(%d, %d, %d)\n",wlevel, wwidth, n_slice+1);
	_create_display_image_of_n(wlevel,wwidth,n_slice);
		glutPostRedisplay();
		return;
		
	  case 's':
		if(wwidth<=2*HUmax+hustep+1){
			wwidth+=hustep;
		}
		printf("(Window Level, Window Width, # of slice)=(%d, %d, %d)\n",wlevel, wwidth, n_slice+1);
	_create_display_image_of_n(wlevel,wwidth,n_slice);
		glutPostRedisplay();
		return;
		
	  case 'x':
		if(wwidth>=-hustep-1){
			wwidth-=hustep;
		}
		printf("(Window Level, Window Width, # of slice)=(%d, %d, %d)\n",wlevel, wwidth, n_slice+1);
	_create_display_image_of_n(wlevel,wwidth,n_slice);
		glutPostRedisplay();
		return;
				
	  case 'f':
		hustep+=1;printf("Changing the HU step for key tapping in window processing: %d\n",hustep);
		return;
		
	  case 'v':
		hustep-=1;printf("Changing the HU step for key tapping in window processing: %d\n",hustep);
		return;
	
	  case 'n':
		if(n_slice<Slice-slabdepth-1){n_slice+=1;}printf("The slice #: %d\n",n_slice+1);
//		_create_display_image_of_n(wlevel,wwidth,n_slice);
		vol2mip(n_slice, slabdepth, wlevel, wwidth);
		glutPostRedisplay();
		return;
		
	  case 'N':
		n_slice=Slice-slabdepth-1;printf("The slice #: %d\n",n_slice+1);
		vol2mip(n_slice, slabdepth, wlevel, wwidth);
		glutPostRedisplay();
		return;

	  case 'b':
		if(n_slice>0){n_slice-=1;}printf("The slice #: %d\n",n_slice+1);
//		_create_display_image_of_n(wlevel,wwidth,n_slice);
		vol2mip(n_slice, slabdepth, wlevel, wwidth);		
		glutPostRedisplay();
		return;
		
	  case 'B':
		n_slice=0;printf("The slice #: %d\n",n_slice+1);
		vol2mip(n_slice, slabdepth, wlevel, wwidth);		
		glutPostRedisplay();
		return;

	  case '<':
		if(slabdepth>1){slabdepth-=1;}printf("the slab depth: %d\n",slabdepth);
		vol2mip(n_slice, slabdepth, wlevel, wwidth);
		glutPostRedisplay();
		return;
		
	  case '>':
		if(slabdepth<Slice){slabdepth+=1;}printf("the slab depth: %d\n",slabdepth);
		vol2mip(n_slice, slabdepth, wlevel, wwidth);
		glutPostRedisplay();
		return;
		
	  case 't':
		n_slice=0;
		printf("Back to Top Slice.\n(Window Level, Window Width, # of slice)=(%d, %d, %d)\n",wlevel, wwidth, n_slice+1);
	_create_display_image_of_n(wlevel,wwidth,n_slice);
		glutPostRedisplay();
		return;
		
	  case 'r':
		wwidth=orgww; wlevel=orgwl;n_slice=0;
		printf("Reset the Window Level & Width and Back to Top Slice.\n(Window Level, Window Width, # of slice)=(%d, %d, %d)\n",wlevel, wwidth, n_slice+1);
	_create_display_image_of_n(wlevel,wwidth,n_slice);
		glutPostRedisplay();
		return;
		
	  case 'e':
		wwidth=orgww; wlevel=orgwl;n_slice=Slice-1;
		printf("Reset the Window Level & Width and jump to end slice.\n(Window Level, Window Width, # of slice)=(%d, %d, %d)\n",wlevel, wwidth, n_slice+1);
	_create_display_image_of_n(wlevel,wwidth,n_slice);
		glutPostRedisplay();
		return;
		
	  case 'w':
		n_slice=Slice-1;
		printf("Jump to end slice.\n(Window Level, Window Width, # of slice)=(%d, %d, %d)\n",wlevel, wwidth, n_slice+1);
	_create_display_image_of_n(wlevel,wwidth,n_slice);
		glutPostRedisplay();
		return;

	  case 'l':
		wwidth=CTpulWW; wlevel=CTpulWL;
		printf("CT: Pulmonary Setting\n(Window Level, Window Width, # of slice)=(%d, %d, %d)\n",wlevel, wwidth, n_slice+1);
		_create_display_image_of_n(wlevel,wwidth,n_slice);
		glutPostRedisplay();
		return;

	  case 'k':
		wwidth=CTabdWW; wlevel=CTabdWL;
		printf("CT: Abdominal Setting\n(Window Level, Window Width, # of slice)=(%d, %d, %d)\n",wlevel, wwidth, n_slice+1);
		_create_display_image_of_n(wlevel,wwidth,n_slice);
		glutPostRedisplay();
		return;


	  case 'j':
		wwidth=CTbonWW; wlevel=CTbonWL;
		printf("CT: Bone Setting\n(Window Level, Window Width, # of slice)=(%d, %d, %d)\n",wlevel, wwidth, n_slice+1);
		_create_display_image_of_n(wlevel,wwidth,n_slice);
		glutPostRedisplay();
		return;

	  case 'h':
		wwidth=CTbraWW; wlevel=CTbraWL;
		printf("CT: Brain Setting\n(Window Level, Window Width, # of slice)=(%d, %d, %d)\n",wlevel, wwidth, n_slice+1);
		_create_display_image_of_n(wlevel,wwidth,n_slice);
		glutPostRedisplay();
		return;
		
	  case 'p':
		printf("\n\nFile information: %s\n", fname);
		printf("W = %d,", Width);
		printf("H = %d,", Height);
		printf("WL = %d,", wlevel);
		printf("WW = %d\n", wwidth);
		printf("HU offset = %d,", huoffset);
		printf("header size[byte] = %d\n", offset);
		printf(" # of slice = %d, ", Slice);
		printf("Slab depth to MIP [slices] = %d\n\n", slabdepth);
        return;
		
	  case 'q':
        exit(1);

      default:
        return;
    }
}

static void Mouse(int button, int state, int mouseX, int mouseY)
{
    if (state != GLUT_DOWN)
        return;
	printf("(x,y, # of slice) = (%d,%d, %d) HU = %d\n",mouseX,mouseY, n_slice+1,ctdata[n_slice*Width*Height + mouseY * Width + mouseX]);
}


unsigned char win_proc(int value, int level, int width)
{
	int i;
	
	if(value>level+width/2){
		return 255;
	}
	
	if(value<level-width/2){
		return 0;
	}
	
	return (unsigned char)(255.0*(value - (level-width/2))/width);
}

void byte_swap_short(short *in, int width,int height)
/*
  inξ̥Хȡ̥ХȤ촹
  width,heightin礭
*/
{
    short org1,org2;
    short low,hi;
    int i, mat_size;

    mat_size = width * height;

    for (i=0;i<mat_size;i++){
      org1=org2=in[i];
      low=org1<<8;      low = low & 0xff00;
      hi=org2>>8;      hi = hi & 0x00ff;

      in[i]=low|hi;
    }
}

void _create_display_image(int wl, int ww)
{
	int i,j;
	
	for(i=0;i<Width*Height;i++){
		image[i]=win_proc(ctdata[i],wl,ww);
	}
}

void vol2mip(int num, int depth, int wl, int ww)
{
	int n, i, j;
	int img_size,vol_size;
	short max,min;
	
	img_size=Width*Height;
	
	max=-30000; min=30000;
	img_size=Width*Height;vol_size=img_size*Slice;
	
	for(i=num*img_size,j=0;i<(num+1)*img_size;i++,j++){
		maximg[j]=ctdata[i];
	}
		
	for(n=num;n<num+depth;n++){
		for(i=n*img_size,j=0;i<(n+1)*img_size;i++,j++){
			if(ctdata[i]>maximg[j]){
				maximg[j]=ctdata[i];
			}
		}
	}
	
	for(i=0;i<img_size;i++){
		image[i] = win_proc(maximg[i], wl, ww);
	}
}

void _create_display_image_of_n(int wl, int ww, int num)
{
	int i,j,k;
	int start_byte, end_byte,num_of_display_slice;
	
	num_of_display_slice = 1;
	start_byte = num*Width*Height;
	end_byte = start_byte + Width*Height*num_of_display_slice;
	
	for(i=start_byte,k=0;i<end_byte;i++,k++){
		image[k]=win_proc(ctdata[i],wl,ww);
//		printf("%d",ctdata[i]);
	}
}

int main(int argc, char **argv){

  GLenum type;

  _get_args(argc, argv);
  ctdata=(short *)calloc(Width*Height*Slice,sizeof(short));
  maximg=(short *)calloc(Width*Height,sizeof(short));
  minimg=(short *)calloc(Width*Height,sizeof(short));
  image=(unsigned char *)calloc(Width*Height,sizeof(char));
	
	_read_image(argv[1]);
	if(swap==0){byte_swap_short(ctdata, Width*Height, Slice);}
	
	//_create_display_image_of_n(wlevel,wwidth,n_slice);
	vol2mip(n_slice, slabdepth, wlevel, wwidth);
	
  glutInit(&argc, argv);

  type = GLUT_RGB;
  type |= (doubleBuffer) ? GLUT_DOUBLE : GLUT_SINGLE;
  glutInitDisplayMode(type);
  glutInitWindowSize(Width*Scale,Height*Scale);
  /*  glutInitWindowPosition(100, 100);*/
  glutCreateWindow(argv[0]);
  glutKeyboardFunc(Key);
	glutMouseFunc(Mouse);
	
  _GL_init();
  glutDisplayFunc(_GL_display);
  glutReshapeFunc(_GL_reshape);
  glutMainLoop();

  free(image);
  return 0; 
}
